from __future__ import annotations

import jax
import jax.numpy as jnp

from dilax.model import Model, Result
from dilax.optimizer import JaxOptimizer
from dilax.parameter import Parameter, compose, lnN, modifier, shape, unconstrained


class SPlusBModel(Model):
    def __call__(
        self,
        processes: dict,
        parameters: dict[str, jax.Array],
    ) -> Result:
        res = Result()

        mu_modifier = modifier(
            name="mu", parameter=parameters["mu"], effect=unconstrained()
        )
        res.add(
            process="signal",
            expectation=mu_modifier(processes[("signal", "nominal")]),
        )

        bkg1_modifier = compose(
            modifier(name="lnN1", parameter=parameters["norm1"], effect=lnN(0.1)),
            modifier(
                name="shape1_bkg1",
                parameter=parameters["shape1"],
                effect=shape(
                    up=processes[("background1", "shape_up")],
                    down=processes[("background1", "shape_down")],
                ),
            ),
        )
        res.add(
            process="background1",
            expectation=bkg1_modifier(processes[("background1", "nominal")]),
        )

        bkg2_modifier = compose(
            modifier(name="lnN2", parameter=parameters["norm2"], effect=lnN(0.05)),
            modifier(
                name="shape1_bkg2",
                parameter=parameters["shape1"],
                effect=shape(
                    up=processes[("background2", "shape_up")],
                    down=processes[("background2", "shape_down")],
                ),
            ),
        )
        res.add(
            process="background2",
            expectation=bkg2_modifier(processes[("background2", "nominal")]),
        )
        return res


def create_model():
    processes = {
        ("signal", "nominal"): jnp.array([3]),
        ("background1", "nominal"): jnp.array([10]),
        ("background2", "nominal"): jnp.array([20]),
        ("background1", "shape_up"): jnp.array([12]),
        ("background1", "shape_down"): jnp.array([8]),
        ("background2", "shape_up"): jnp.array([23]),
        ("background2", "shape_down"): jnp.array([19]),
    }
    parameters = {
        "mu": Parameter(value=jnp.array([1.0]), bounds=(0.0, jnp.inf)),
        "norm1": Parameter(value=jnp.array([0.0])),
        "norm2": Parameter(value=jnp.array([0.0])),
        "shape1": Parameter(value=jnp.array([0.0])),
    }

    # return model
    return SPlusBModel(processes=processes, parameters=parameters)


model = create_model()

init_values = model.parameter_values
observation = jnp.array([37])
asimov = model.evaluate().expectation()


# create optimizer (from `jaxopt`)
optimizer = JaxOptimizer.make(
    name="LBFGS",
    settings={"maxiter": 5, "jit": True, "unroll": True},
)



from __future__ import annotations

import abc

import equinox as eqx
import jax
import jax.numpy as jnp

from dilax.pdf import Flat, Gauss, HashablePDF, Poisson
from dilax.util import as1darray

__all__ = [
    "Parameter",
    "Effect",
    "unconstrained",
    "gauss",
    "lnN",
    "poisson",
    "shape",
    "modifier",
    "staterror",
    "compose",
]


def __dir__():
    return __all__


class Parameter(eqx.Module):
    value: jax.Array = eqx.field(converter=as1darray)
    bounds: tuple[jax.Array, jax.Array] = eqx.field(
        static=True, converter=lambda x: tuple(map(as1darray, x))
    )
    constraints: set[HashablePDF] = eqx.field(static=True)

    def __init__(
        self,
        value: jax.Array,
        bounds: tuple[jax.Array, jax.Array] = (as1darray(-jnp.inf), as1darray(jnp.inf)),
    ) -> None:
        self.value = value
        self.bounds = bounds
        self.constraints: set[HashablePDF] = set()

    def update(self, value: jax.Array) -> Parameter:
        return eqx.tree_at(lambda t: t.value, self, value)

    @property
    def boundary_penalty(self) -> jax.Array:
        return jnp.where(
            (self.value < self.bounds[0]) | (self.value > self.bounds[1]),
            jnp.inf,
            0,
        )


class Effect(eqx.Module):
    @property
    @abc.abstractmethod
    def constraint(self) -> HashablePDF:
        ...

    @abc.abstractmethod
    def scale_factor(self, parameter: Parameter, sumw: jax.Array) -> jax.Array:
        ...


class unconstrained(Effect):
    @property
    def constraint(self) -> HashablePDF:
        return Flat()

    def scale_factor(self, parameter: Parameter, sumw: jax.Array) -> jax.Array:
        return parameter.value


DEFAULT_EFFECT = unconstrained()


class gauss(Effect):
    width: jax.Array = eqx.field(static=True, converter=as1darray)

    def __init__(self, width: jax.Array) -> None:
        self.width = width

    @property
    def constraint(self) -> HashablePDF:
        return Gauss(mean=0.0, width=1.0)

    def scale_factor(self, parameter: Parameter, sumw: jax.Array) -> jax.Array:
        # gx = Gauss(mean=1.0, width=self.width)  # type: ignore[arg-type]
        # g1 = Gauss(mean=1.0, width=1.0)
        # return gx.inv_cdf(g1.cdf(parameter.value + 1))
        return (parameter.value * self.width) + 1  # fast analytical solution


class shape(Effect):
    up: jax.Array = eqx.field(converter=as1darray)
    down: jax.Array = eqx.field(converter=as1darray)

    def __init__(
        self,
        up: jax.Array,
        down: jax.Array,
    ) -> None:
        self.up = up  # +1 sigma
        self.down = down  # -1 sigma

    @eqx.filter_jit
    def vshift(self, sf: jax.Array, sumw: jax.Array) -> jax.Array:
        factor = sf
        dx_sum = self.up + self.down - 2 * sumw
        dx_diff = self.up - self.down

        # taken from https://github.com/nsmith-/jaxfit/blob/8479cd73e733ba35462287753fab44c0c560037b/src/jaxfit/roofit/combine.py#L173C6-L192
        _asym_poly = jnp.array([3.0, -10.0, 15.0, 0.0]) / 8.0

        abs_value = jnp.abs(factor)
        return 0.5 * (
            dx_diff * factor
            + dx_sum
            * jnp.where(
                abs_value > 1.0,
                abs_value,
                jnp.polyval(_asym_poly, factor * factor),
            )
        )

    @property
    def constraint(self) -> HashablePDF:
        return Gauss(mean=0.0, width=1.0)

    def scale_factor(self, parameter: Parameter, sumw: jax.Array) -> jax.Array:
        sf = parameter.value
        # clip, no negative values are allowed
        return jnp.maximum((sumw + self.vshift(sf=sf, sumw=sumw)) / sumw, 0.0)


class lnN(Effect):
    width: jax.Array | tuple[jax.Array, jax.Array] = eqx.field(static=True)

    def __init__(
        self,
        width: jax.Array | tuple[jax.Array, jax.Array],
    ) -> None:
        self.width = width

    def scale(self, parameter: Parameter) -> jax.Array:
        if isinstance(self.width, tuple):
            down, up = self.width
            scale = jnp.where(parameter.value > 0, up, down)
        else:
            scale = self.width
        return scale

    @property
    def constraint(self) -> HashablePDF:
        return Gauss(mean=0.0, width=1.0)

    def scale_factor(self, parameter: Parameter, sumw: jax.Array) -> jax.Array:
        # width = self.scale(parameter=parameter)
        # g1 = Gauss(mean=1.0, width=1.0)
        # gx = Gauss(mean=jnp.exp(parameter.value), width=width)  # type: ignore[arg-type]
        # return gx.inv_cdf(g1.cdf(parameter.value + 1))
        return jnp.exp(
            parameter.value * self.scale(parameter=parameter)
        )  # fast analytical solution


class poisson(Effect):
    lamb: jax.Array = eqx.field(static=True, converter=as1darray)

    def __init__(self, lamb: jax.Array) -> None:
        self.lamb = lamb

    @property
    def constraint(self) -> HashablePDF:
        return Poisson(lamb=self.lamb)

    def scale_factor(self, parameter: Parameter, sumw: jax.Array) -> jax.Array:
        return parameter.value + 1


class ModifierBase(eqx.Module):
    @abc.abstractmethod
    def scale_factor(self, sumw: jax.Array) -> jax.Array:
        ...

    def __call__(self, sumw: jax.Array) -> jax.Array:
        return jnp.atleast_1d(self.scale_factor(sumw=sumw)) * sumw


class modifier(ModifierBase):
    """
    Create a new modifier for a given parameter and penalty.

    Example:

    .. code-block:: python

        import jax.numpy as jnp
        from dilax.parameter import modifier, Parameter, unconstrained, lnN, poisson, shape

        mu = Parameter(value=1.1, bounds=(0, 100))
        norm = Parameter(value=0.0, bounds=(-jnp.inf, jnp.inf))

        # create a new parameter and a penalty
        modify = modifier(name="mu", parameter=mu, effect=unconstrained())

        # apply the modifier
        modify(jnp.array([10, 20, 30]))
        # -> Array([11., 22., 33.], dtype=float32, weak_type=True),

        # lnN effect
        modify = modifier(name="norm", parameter=norm, effect=lnN(0.2))
        modify(jnp.array([10, 20, 30]))

        # poisson effect
        hist = jnp.array([10, 20, 30])
        modify = modifier(name="norm", parameter=norm, effect=poisson(hist))
        modify(jnp.array([10, 20, 30]))

        # shape effect
        up = jnp.array([12, 23, 35])
        down = jnp.array([8, 19, 26])
        modify = modifier(name="norm", parameter=norm, effect=shape(up, down))
        modify(jnp.array([10, 20, 30]))
    """

    name: str
    parameter: Parameter
    effect: Effect

    def __init__(
        self, name: str, parameter: Parameter, effect: Effect = DEFAULT_EFFECT
    ) -> None:
        self.name = name
        self.parameter = parameter
        self.effect = effect
        self.parameter.constraints.add(self.effect.constraint)

    def scale_factor(self, sumw: jax.Array) -> jax.Array:
        return self.effect.scale_factor(parameter=self.parameter, sumw=sumw)


class staterror(ModifierBase):
    """
    Create a staterror (barlow-beeston) modifier which acts on each bin with a different _underlying_ modifier.

    Example:

    .. code-block:: python

        import jax.numpy as jnp
        from dilax.parameter import modifier, Parameter, unconstrained, lnN, poisson, shape

        hist = jnp.array([10, 20, 30])

        p1 = Parameter(value=1.0)
        p2 = Parameter(value=0.0)
        p3 = Parameter(value=0.0)

        # all bins with bin content below 10 (threshold) are treated as poisson, else gauss
        modify = staterror(
            parameters=[p1, p2, p3],
            sumw=hist,
            sumw2=hist,
            threshold=10.0,
        )
        modify(hist)
        # -> Array([13.162277, 20.      , 30.      ], dtype=float32)

        fast_modify = eqx.filter_jit(modify)
    """

    name: str = "staterror"
    parameters: list[Parameter]
    sumw: jax.Array
    sumw2: jax.Array
    sumw2sqrt: jax.Array
    widths: jax.Array
    mask: jax.Array
    threshold: float

    def __init__(
        self,
        parameters: list[Parameter],
        sumw: jax.Array,
        sumw2: jax.Array,
        threshold: float,
    ) -> None:
        assert len(parameters) == len(sumw2) == len(sumw)

        self.parameters = parameters
        self.sumw = sumw
        self.sumw2 = sumw2
        self.sumw2sqrt = jnp.sqrt(sumw2)
        self.threshold = threshold

        # calculate width
        self.widths = self.sumw2sqrt / self.sumw

        # store if sumw is below threshold
        self.mask = self.sumw < self.threshold

        for i, param in enumerate(self.parameters):
            effect = poisson(self.sumw[i]) if self.mask[i] else gauss(self.widths[i])
            param.constraints.add(effect.constraint)

    def scale_factor(self, sumw: jax.Array) -> jax.Array:
        from functools import partial

        assert len(sumw) == len(self.parameters) == len(self.sumw2)

        values = jnp.concatenate([param.value for param in self.parameters])
        idxs = jnp.arange(len(sumw))

        # sumw where mask (poisson) else widths (gauss)
        _widths = jnp.where(self.mask, self.sumw, self.widths)

        def _mod(
            value: jax.Array,
            width: jax.Array,
            idx: jax.Array,
            effect: Effect,
        ) -> jax.Array:
            return effect(width).scale_factor(
                parameter=Parameter(value=value),
                sumw=sumw[idx],
            )[0]

        _poisson_mod = partial(_mod, effect=poisson)
        _gauss_mod = partial(_mod, effect=gauss)

        # where mask use poisson else gauss
        return jnp.where(
            self.mask,
            jax.vmap(_poisson_mod)(values, _widths, idxs),
            jax.vmap(_gauss_mod)(values, _widths, idxs),
        )


class compose(ModifierBase):
    """
    Composition of multiple modifiers, i.e.: `(f ∘ g ∘ h)(hist) = f(hist) * g(hist) * h(hist)`
    It behaves like a single modifier, but it is composed of multiple modifiers; it can be arbitrarly nested.

    Example:

    .. code-block:: python

        from dilax.parameter import modifier, compose, Parameter, unconstrained, lnN

        mu = Parameter(value=1.1, bounds=(0, 100))
        sigma = Parameter(value=0.1, bounds=(-100, 100))

        # create a new parameter and a composition of modifiers
        composition = compose(
            modifier(name="mu", parameter=mu),
            modifier(name="sigma1", parameter=sigma, effect=lnN(0.1)),
        )

        # apply the composition
        composition(jnp.array([10, 20, 30]))

        # nest compositions
        composition = compose(
            composition,
            modifier(name="sigma2", parameter=sigma, effect=lnN(0.2)),
        )

        # jit
        eqx.filter_jit(composition)(jnp.array([10, 20, 30]))
    """

    modifiers: tuple[modifier, ...]
    names: list[str] = eqx.field(static=True)

    def __init__(self, *modifiers: modifier) -> None:
        self.modifiers = modifiers

        # set names
        self.names = []
        for m in range(len(self)):
            modifier = self.modifiers[m]
            if isinstance(modifier, compose):
                self.names.extend(modifier.names)
            else:
                self.names.append(modifier.name)

        # check for duplicate names
        duplicates = [name for name in self.names if self.names.count(name) > 1]
        if duplicates:
            msg = f"Modifier need to have unique names, got: {duplicates}"
            raise ValueError(msg)

    def __len__(self) -> int:
        return len(self.modifiers)

    def scale_factors(self, sumw: jax.Array) -> dict[str, jax.Array]:
        sfs = {}
        for m in range(len(self)):
            modifier = self.modifiers[m]
            if isinstance(modifier, compose):
                sfs.update(modifier.scale_factors(sumw=sumw))
            else:
                sf = jnp.atleast_1d(modifier.scale_factor(sumw=sumw))
                sfs[modifier.name] = jnp.broadcast_to(sf, sumw.shape)
        return sfs

    def scale_factor(self, sumw: jax.Array) -> jax.Array:
        sfs = jnp.stack(list(self.scale_factors(sumw=sumw).values()))
        # calculate the product in log-space for numerical precision
        return jnp.exp(jnp.sum(jnp.log(sfs), axis=0))



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, cast

import equinox as eqx
import jax
import jax.numpy as jnp

from dilax.parameter import Parameter
from dilax.util import Sentinel, _NoValue

__all__ = ["Result", "Model"]


def __dir__():
    return __all__


class Result(eqx.Module):
    expectations: dict[str, jax.Array]

    def __init__(self) -> None:
        self.expectations = {}

    def add(self, process: str, expectation: jax.Array) -> Result:
        self.expectations[process] = expectation
        return self

    def expectation(self) -> jax.Array:
        return cast(jax.Array, sum(jax.tree_util.tree_leaves(self.expectations)))


class Model(eqx.Module):
    """
    A model describing nuisance parameters, templates (histograms), and how they interact.
    It is requires to implement the `evaluate` method, which returns an `EvaluationResult` object.

    Example:

    .. code-block:: python

        import jax
        import jax.numpy as jnp
        import equinox as eqx

        from dilax.model import Model, Result
        from dilax.parameter import Parameter, lnN, modifier, unconstrained


        # Define a simple model with two processes and two parameters
        class MyModel(Model):
            def __call__(self, processes: dict, parameters: dict[str, Parameter]) -> Result:
                res = Result()

                # signal
                mu_mod = modifier(name="mu", parameter=parameters["mu"], effect=unconstrained())
                res.add(process="signal", expectation=mu_mod(processes["signal"]))

                # background
                bkg_mod = modifier(name="sigma", parameter=parameters["sigma"], effect=lnN(0.2))
                res.add(process="background", expectation=bkg_mod(processes["background"]))
                return res


        # Setup model
        processes = {"signal": jnp.array([10]), "background": jnp.array([50])}
        parameters = {
            "mu": Parameter(value=jnp.array([1.0]), bounds=(0.0, jnp.inf)),
            "sigma": Parameter(value=jnp.array([0.0])),
        }

        model = MyModel(processes=processes, parameters=parameters)

        # evaluate the expectation
        model.evaluate().expectation()
        # -> Array([60.], dtype=float32)

        %timeit model.evaluate().expectation()
        # -> 3.05 ms ± 29.3 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)

        # evaluate the expectation *fast*
        @eqx.filter_jit
        def eval(model) -> jax.Array:
            res = model.evaluate()
            return res.expectation()

        eqx.filter_jit(eval)(model)
        # -> Array([60.], dtype=float32)

        %timeit eqx.filter_jit(eval)(model).block_until_ready()
        # -> 114 µs ± 327 ns per loop (mean ± std. dev. of 7 runs, 10,000 loops each)
    """

    processes: dict
    parameters: dict[str, Parameter]

    def __init__(self, processes: dict, parameters: dict[str, Parameter]) -> None:
        self.processes = processes
        self.parameters = parameters

    @property
    def parameter_values(self) -> dict[str, jax.Array]:
        return {key: param.value for key, param in self.parameters.items()}

    def parameter_constraints(self) -> dict[str, jax.Array]:
        constraints = {}
        for name, param in self.parameters.items():
            # skip if the parameter was not used / has no constraint
            if not param.constraints:
                continue
            if not len(param.constraints) <= 1:
                msg = f"More than one constraint per parameter is not allowed. Got: {param.constraint}"
                raise ValueError(msg)
            constraint = next(iter(param.constraints))
            constraints[name] = constraint.logpdf(param.value)
        return constraints

    def update(
        self,
        processes: dict | Sentinel = _NoValue,
        values: dict[str, jax.Array] | Sentinel = _NoValue,
    ) -> Model:
        if values is _NoValue:
            values = {}
        if processes is _NoValue:
            processes = {}

        if TYPE_CHECKING:
            values = cast(dict[str, jax.Array], values)
            processes = cast(dict, processes)

        # patch original processes with new ones
        new_processes = {}
        for key, old_process in self.processes.items():
            if key in processes:
                new_process = processes[key]
                new_processes[key] = new_process
            else:
                new_processes[key] = old_process

        # patch original parameters with new ones
        new_parameters = {}
        for key, old_parameter in self.parameters.items():
            if key in values:
                new_parameter = old_parameter.update(value=values[key])
                new_parameters[key] = new_parameter
            else:
                new_parameters[key] = old_parameter

        return eqx.tree_at(
            lambda t: (t.processes, t.parameters), self, (new_processes, new_parameters)
        )

    def nll_boundary_penalty(self) -> jax.Array:
        penalty = jnp.array([0.0])

        for param in self.parameters.values():
            penalty += param.boundary_penalty

        return penalty

    @abc.abstractmethod
    def __call__(self, processes: dict, parameters: dict[str, Parameter]) -> Result:
        ...

    def evaluate(self) -> Result:
        # evaluate the model with its current state
        return self(self.processes, self.parameters)



"""
dilax: Differentiable (binned) likelihoods in JAX.
"""
from __future__ import annotations

__author__ = "Peter Fackeldey"
__email__ = "peter.fackeldey@rwth-aachen.de"
__copyright__ = "Copyright 2023, Peter Fackeldey"
__credits__ = ["Peter Fackeldey"]
__contact__ = "https://github.com/pfackeldey/dilax"
__license__ = "BSD-3-Clause"
__status__ = "Development"
__version__ = "0.1.5"


# expose public API

__all__ = [
    "ipy_util",
    "likelihood",
    "model",
    "optimizer",
    "parameter",
    # "pdf", # this should not be needed in public API
    "util",
    "__version__",
]


def __dir__():
    return __all__


from dilax import (  # noqa: E402
    ipy_util,
    likelihood,
    model,
    optimizer,
    parameter,
    # pdf,  # this should not be needed in public API
    util,
)
